! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: esources_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Selects EPIC sources for creation of spectra and time series.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module esources_mod

  use dal
  use param  
  use errorHandling

  use dal_aux

  implicit none 

  private 

  type :: instWeightType
    integer(int16) :: idInst
    real(single)   :: weight
  end type instWeightType

  ! Command-line parameters:
  character(256), allocatable ::&
    eventSetNames(:)
  character(256) ::&
    srcmatchSourceSetTabName='',&
    outputSourceSetName=''
  real(single) ::&
    minEpicCounts=0.0,&
    minDetml=0.0,&
    minMaskFrac=0.0
  integer ::&
    maxNumHexSrcChars=0 ! [1:7]
  type(instWeightType), allocatable ::&
    weights(:)
  logical ::&
    setFlagsInInputList=.false.

  ! Other globals:
  character(256) :: srcmatchSourceSetName
  character(20)  :: srcmatchSourceTabName

  type :: instSrcDataType
    logical      :: sctsIsNull, isAboveThre
    real(single) :: scts, detml, maskFrac
    real(double) :: ontime
  end type instSrcDataType

  integer, parameter :: LENGTHOFINSTFLAG = 12

  type :: srcDataType
    real(single)          :: scts, binWidth
    logical               :: sctsIsNull
    type(instSrcDataType) :: PN, M1, M2
    character(7)          :: srcNumHex 
    integer(int32)        :: bitFlags, srcNum, mlIdSrc
    real(double)          :: ra, dec, raCorr, decCorr
    character(LENGTHOFINSTFLAG) :: m1Flag, m2Flag, pnFlag, epFlag
  end type srcDataType

  integer, parameter :: pnFlagBit = 0, m1FlagBit = 1, m2FlagBit = 2

  public esources

contains 

  !---------------------------------------------------------------------
  subroutine esources
  !---------------------------------------------------------------------

    implicit none

    real(single), parameter :: pnMinSctsPerBin = 18.0,&
                               m1MinSctsPerBin =  5.0,&
                               m2MinSctsPerBin =  5.0

    integer :: maxSrcNum, i, numSources
    type(srcDataType), pointer :: srcData(:)
    real(single) :: usedWeight, binWidthReal, totalWeight
    real(double) :: startTime, stopTime
    character(1023) :: tmpc

!logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    maxSrcNum = 1 ! default
    do i = 1, maxNumHexSrcChars
      maxSrcNum = maxSrcNum * 16
    end do
    maxSrcNum = maxSrcNum - 1

    totalWeight = sum(weights%weight)

    call readSourceListData(srcData)
    numSources = size(srcData)

    do i = 1, numSources
      ! Compute total epic counts:
      !  Definition of total epic counts is that if counts of all
      !  the instruments exist then PN+M1+M2; however if not,
      !  a missing one(s) has to be compensated.
      !  See some 35 lines below for the concrete formula.


      ! Set up a set of logical constants for each source.
      if (  srcData(i)%PN%detml    >  minDetml&           ! default 15
      .and. srcData(i)%PN%maskFrac >= minMaskFrac)&       ! default 0.5
      then
        srcData(i)%PN%isAboveThre = .true.
      else
        srcData(i)%PN%isAboveThre = .false.
      end if

      if (  srcData(i)%M1%detml    >  minDetml&           ! default 15
      .and. srcData(i)%M1%maskFrac >= minMaskFrac)&       ! default 0.5
      then
        srcData(i)%M1%isAboveThre = .true.
      else
        srcData(i)%M1%isAboveThre = .false.
      end if

      if (  srcData(i)%M2%detml    >  minDetml&           ! default 15
      .and. srcData(i)%M2%maskFrac >= minMaskFrac)&       ! default 0.5
      then
        srcData(i)%M2%isAboveThre = .true.
      else
        srcData(i)%M2%isAboveThre = .false.
      end if


      ! Start calculation:
      usedWeight = 0.0

      if (.not.srcData(i)%PN%sctsIsNull .and. srcData(i)%PN%isAboveThre) then
        srcData(i)%scts = srcData(i)%scts + srcData(i)%PN%scts
        usedWeight      = usedWeight      + weights(1)%weight
      end if

      if (.not.srcData(i)%M1%sctsIsNull .and. srcData(i)%M1%isAboveThre) then
        srcData(i)%scts = srcData(i)%scts + srcData(i)%M1%scts
        usedWeight      = usedWeight      + weights(2)%weight
      end if

      if (.not.srcData(i)%M2%sctsIsNull .and. srcData(i)%M2%isAboveThre) then
        srcData(i)%scts = srcData(i)%scts + srcData(i)%M2%scts
        usedWeight      = usedWeight      + weights(3)%weight
      end if

      if (usedWeight == 0.0) then
        srcData(i)%sctsIsNull = .true.
      else
        srcData(i)%scts = srcData(i)%scts * totalWeight / usedWeight
        ! EPIC-counts = Sum(CTS_PN,CTS_M1,CTS_M2)
        !                * Sum(Weight_PN+W_M1+W_M2)/Sum(Used(W_PN,W_M1,W_M2))
      end if
      ! Note that this formula is arithmetically identical to that used
      ! in ACS's script, except for the case in which SCTS is null only
      ! in one of the mos cameras. In this aberrant case the results
      ! of the two methods are equal on average, but the uncertainties
      ! are different.

      if (srcData(i)%scts >= minEpicCounts) then ! default 500
        if (srcData(i)%PN%isAboveThre)&
          srcData(i)%bitFlags = ibset(srcData(i)%bitFlags, pnFlagBit)

        if (srcData(i)%M1%isAboveThre)&
          srcData(i)%bitFlags = ibset(srcData(i)%bitFlags, m1FlagBit)

        if (srcData(i)%M2%isAboveThre)&
          srcData(i)%bitFlags = ibset(srcData(i)%bitFlags, m2FlagBit)
      end if


      ! Calculate bin width for each source (should give identical results
      ! to ACS):

      if (     .not.srcData(i)%PN%sctsIsNull .and. (srcData(i)%PN%scts > 0.0) .and.&
               srcData(i)%PN%isAboveThre .and. (srcData(i)%PN%ontime > 0.0) ) then
        binWidthReal = floor(pnMinSctsPerBin * srcData(i)%PN%ontime&
          / srcData(i)%PN%scts)

      else if (.not.srcData(i)%M2%sctsIsNull .and. (srcData(i)%M2%scts > 0.0) .and.&
               srcData(i)%M2%isAboveThre .and. (srcData(i)%M2%ontime > 0.0) ) then
               ! MOS2 is given a higher priority than MOS1
        binWidthReal = floor(m2MinSctsPerBin * srcData(i)%M2%ontime&
          / srcData(i)%M2%scts)

      else if (.not.srcData(i)%M1%sctsIsNull .and. (srcData(i)%M1%scts > 0.0) .and.&
               srcData(i)%M1%isAboveThre .and. (srcData(i)%M1%ontime > 0.0) ) then
        binWidthReal = floor(m1MinSctsPerBin * srcData(i)%M1%ontime&
          / srcData(i)%M1%scts)

      else
        ! This source is not "significant" for the given threshold (or ontime).
        srcData(i)%bitFlags = 0		! Skip this row in writing FITS.
      end if

      srcData(i)%binWidth = 10.0d0 * (1 + floor(binWidthReal/10.0))
      ! NOTE that this formula implies that there will always be LESS than
      ! the desired counts per bin.

      if (srcData(i)%binWidth <= 0.0) then
        write (tmpc,'("Skip writing the source number: ",I4)') i
        call warning('binWidthTooSmall', tmpc)
        srcData(i)%bitFlags = 0
      end if

      ! srcData(i)%srcNumHex = intToHex(srcData(i)%srcNum)  ! function removed
      write(srcData(i)%srcNumHex, '(Z0)') srcData(i)%srcNum

      if (srcData(i)%bitFlags .ne. 0 .and. srcData(i)%srcNum > maxSrcNum) then
        call warning('srcNumTooLarge', "")
        srcData(i)%bitFlags = 0
      end if
    end do

    call getStartAndStopTimes(startTime, stopTime)

    call writeSourceListData(srcData, startTime, stopTime)

    if (setFlagsInInputList .and. any(srcData%bitFlags > 0))&
      call setFlags((srcData%bitFlags > 0))

    deallocate(srcData)

  end subroutine esources

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    integer :: numEventLists, i, numWeights

    !-------------------------------------------------------------------
    ! Start

    numEventLists = parameterCount('eventsets')
if (numEventLists <= 0) call error('noEventListsSupplied',"")
    allocate(eventSetNames(numEventLists))
    do i = 1, numEventLists
      eventSetNames(i) = stringParameter('eventsets',i-1)
    end do

    srcmatchSourceSetTabName = stringParameter('srctab')
    outputSourceSetName      = stringParameter('outsrcset')
    minEpicCounts            = realParameter(  'mintotalcts')
    minDetml                 = realParameter(  'mindetml')
    minMaskFrac              = realParameter(  'minmaskfrac')
    maxNumHexSrcChars        = intParameter(   'numhexchar') ! [1:7]

    numWeights = parameterCount('instweights')
if (numWeights .ne. 3) call error('badNumWeights',"")
    allocate(weights(numWeights))
    do i = 1, numWeights
      weights(i)%idInst = i
      weights(i)%weight = realParameter('instweights', i-1)
    end do

    setFlagsInInputList = booleanParameter('setflags')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine readSourceListData(srcData)
  !---------------------------------------------------------------------
  ! Note: the calling routine should deallocate the pointer srcData after use.

    implicit none

    type(srcDataType), pointer :: srcData(:)

!    character(256) :: setName
!    character(20) :: tabName
    type(DataSetT) :: set
    type(TableT) :: tab
    integer :: numSources, row
    type(ColumnT) ::&
      pnSctsCol, m1SctsCol, m2SctsCol, pnOntimeCol, m1OntimeCol, m2OntimeCol

    !-------------------------------------------------------------------
    ! Start

    call splitSetTabName(srcmatchSourceSetTabName, srcmatchSourceSetName&
      , srcmatchSourceTabName) ! in ssclib/dal_aux.

    set = dataSet(srcmatchSourceSetName, READ)
    tab = table(set, srcmatchSourceTabName)
    numSources = numberOfRows(tab)
    allocate(srcData(numSources))

    if (numSources > 0) then
      call initializeSrcDataType(srcData)

      ! The call readColDataToFixed is in ssclib/dal_aux.
      call readColDataToFixed(tab, 'SRC_NUM',  srcData%srcNum)
      call readColDataToFixed(tab, 'RA',       srcData%ra)
      call readColDataToFixed(tab, 'DEC',      srcData%dec)
      call readColDataToFixed(tab, 'RA_CORR',  srcData%raCorr)
      call readColDataToFixed(tab, 'DEC_CORR', srcData%decCorr)

      pnSctsCol = column(tab, 'PN_CTS')
      m1SctsCol = column(tab, 'M1_CTS')
      m2SctsCol = column(tab, 'M2_CTS')
      pnOntimeCol = column(tab, 'PN_ONTIME')
      m1OntimeCol = column(tab, 'M1_ONTIME')
      m2OntimeCol = column(tab, 'M2_ONTIME')

      do row = 0, numSources-1
        ! if (isNull(pnSctsCol, row) .or. isNull(pnOntimeCol, row))&
        if (isNull(pnSctsCol, row))&
          srcData(row+1)%PN%sctsIsNull = .true.
        if (isNull(m1SctsCol, row))&
          srcData(row+1)%M1%sctsIsNull = .true.
        if (isNull(m2SctsCol, row))&
          srcData(row+1)%M2%sctsIsNull = .true.
      end do

      ! The call readColDataToFixed is in ssclib/dal_aux.
      call readColDataToFixed(pnSctsCol,          srcData%PN%scts)
      call readColDataToFixed(tab, 'PN_DET_ML',   srcData%PN%detml)
      call readColDataToFixed(tab, 'PN_MASKFRAC', srcData%PN%maskFrac)
      call readColDataToFixed(pnOntimeCol,   srcData%PN%ontime)

      call readColDataToFixed(m1SctsCol,          srcData%M1%scts)
      call readColDataToFixed(tab, 'M1_DET_ML',   srcData%M1%detml)
      call readColDataToFixed(tab, 'M1_MASKFRAC', srcData%M1%maskFrac)
      call readColDataToFixed(m1OntimeCol,   srcData%M1%ontime)

      call readColDataToFixed(m2SctsCol,          srcData%M2%scts)
      call readColDataToFixed(tab, 'M2_DET_ML',   srcData%M2%detml)
      call readColDataToFixed(tab, 'M2_MASKFRAC', srcData%M2%maskFrac)
      call readColDataToFixed(m2OntimeCol,   srcData%M2%ontime)

      srcData%pnFlag = ''
      if (hasColumn(tab, 'PN_FLAG')) then
        call readColDataToFixed(tab, 'PN_FLAG', srcData%pnFlag)
      end if
      srcData%m1Flag = ''
      if (hasColumn(tab, 'M1_FLAG')) then
        call readColDataToFixed(tab, 'M1_FLAG', srcData%m1Flag)
      end if
      srcData%m2Flag = ''
      if (hasColumn(tab, 'M2_FLAG')) then
        call readColDataToFixed(tab, 'M2_FLAG', srcData%m2Flag)
      end if
      srcData%epFlag = ''
      if (hasColumn(tab, 'EP_FLAG')) then
        call readColDataToFixed(tab, 'EP_FLAG', srcData%epFlag)
      end if

      if (hasColumn(tab, 'EP_1_ML_ID')) then
        call readColDataToFixed(tab, 'EP_1_ML_ID', srcData%mlIdSrc)
      else
        call warning('noEp1MlIdCol',"")
      end if

      call release(m2OntimeCol)
      call release(m1OntimeCol)
      call release(pnOntimeCol)
      call release(m2SctsCol)
      call release(m1SctsCol)
      call release(pnSctsCol)
    end if ! numSources > 0

    call release(tab)
    call release(set)

  end subroutine readSourceListData

  !---------------------------------------------------------------------
  subroutine initializeSrcDataType(srcData)
  !---------------------------------------------------------------------

    implicit none

    type(srcDataType), intent(inout) :: srcData(:)

    !-------------------------------------------------------------------
    ! Start

    srcData%scts       = 0.0
    srcData%binWidth   = 0.0
    srcData%sctsIsNull = .false.
    srcData%srcNumHex  = ''
    srcData%bitFlags   = 0
    srcData%srcNum     = 0
    srcData%mlIdSrc    = 0
    srcData%ra         = 0.0d0
    srcData%dec        = 0.0d0
    srcData%raCorr     = 0.0d0
    srcData%decCorr    = 0.0d0

    call initializeInstSrcDataType(srcData%PN)
    call initializeInstSrcDataType(srcData%M1)
    call initializeInstSrcDataType(srcData%M2)

  end subroutine initializeSrcDataType

  !---------------------------------------------------------------------
  subroutine initializeInstSrcDataType(instSrcData)
  !---------------------------------------------------------------------

    implicit none

    type(instSrcDataType), intent(inout) :: instSrcData(:)

    !-------------------------------------------------------------------
    ! Start

    instSrcData%sctsIsNull = .false.
    instSrcData%scts       = 0.0
    instSrcData%detml      = 0.0
    instSrcData%maskFrac   = 0.0
    instSrcData%ontime     = 0.0d0

  end subroutine initializeInstSrcDataType

  !---------------------------------------------------------------------
  subroutine getStartAndStopTimes(startTime, stopTime)
  !---------------------------------------------------------------------
  ! Reads TSTART and TSTOP kwds from all event list headers and computes the OR of these periods.
  ! 
  ! It used to be stated as
  ! | This ORed interval gets written as TLMIN and TLMAX kwds to the time column of each event list and is also written to the output src list.
  ! and so was in the reference manual.  However it has never been done,
  ! at least properly, due to a bug.  Anyhow it is not a desirable feature.
  ! Hence it is officially not supported as of Version 1.4.1.

    implicit none

    real(double), intent(out) :: startTime, stopTime

    integer :: numEventLists, i
    real(double), allocatable :: startTimes(:), stopTimes(:)
    type(DataSetT) :: set
    type(TableT) :: tab
    type(ColumnT) :: col

    !-------------------------------------------------------------------
    ! Start

    numEventLists = size(eventSetNames)
    if (numEventLists <= 0)&
call error('bug1',"This shouldn't happen. Please &
&contact the code developer.")

    allocate(startTimes(numEventLists), stopTimes(numEventLists))
    startTimes = 0.0d0
    stopTimes  = 0.0d0
    do i = 1, numEventLists
      set = dataSet(eventSetNames(i), READ)
      tab = table(set, 'EVENTS')
      startTimes(i) = real64Attribute(tab, 'TSTART')
      stopTimes( i) = real64Attribute(tab, 'TSTOP')
      call release(tab)
      call release(set)
    end do

    startTime = minval(startTimes)
    stopTime  = maxval(stopTimes)
if (startTime >= stopTime) call error('bug2',"Please contact the code developer.")

    !! NOTE: This part was active until Version 1.4 but commented out at 1.4.1 and after.
    !! See the reference manual (in a footnote) for detail.
    !! 
    ! do i = 1, numEventLists
    !   set = dataSet(eventSetNames(1), MODIFY)		! "1" should be "i" (by Masa, 2010/01/26).
    !   tab = table(set, 'EVENTS')
    !   col = column(tab, 'TIME')
    !   call setAttribute(col, 'TLMIN', startTime)
    !   call setAttribute(col, 'TLMAX', stopTime)
    !   call release(col)
    !   call release(tab)
    !   call release(set)
    ! end do

  end subroutine getStartAndStopTimes

  !---------------------------------------------------------------------
  subroutine writeSourceListData(srcData, startTime, stopTime)
  !---------------------------------------------------------------------

    implicit none

    type(srcDataType), intent(in) :: srcData(:)
    real(double),      intent(in) :: startTime, stopTime

    integer :: numSources, numLiveSources, row, i
    type(DataSetT) :: set
    type(TableT) :: tab
    type(ColumnT) ::&
      srcNumCol, raCol, decCol, raCorrCol, decCorrCol, binWidthCol, srcNumHexCol,&
      bitFlagsCol, epicSctsCol, pnSctsCol, m1SctsCol, m2SctsCol, mlIdSrcCol,&
      epFlagCol, pnFlagCol, m1FlagCol, m2FlagCol
    integer(int32), pointer :: srcNumData(:), mlIdSrcData(:)
    real(double), pointer ::&
      raData(:), decData(:), raCorrData(:), decCorrData(:), binWidthData(:)
    integer(int8), pointer :: bitFlagsData(:)
    real(single), pointer :: epicSctsData(:), pnSctsData(:), m1SctsData(:), m2SctsData(:)

    !-------------------------------------------------------------------
    ! Start

    numSources = size(srcData)
    numLiveSources = count(srcData%bitFlags > 0)

    set = dataSet(outputSourceSetName, CREATE)
    tab = addTable(set, 'SRCLIST', numLiveSources)

    call setAttribute(tab, 'PN_P_BIT', pnFlagBit)
    call setAttribute(tab, 'M1_P_BIT', m1FlagBit)
    call setAttribute(tab, 'M2_P_BIT', m2FlagBit)

    call setAttribute(tab, 'TSTART', startTime)
    call setAttribute(tab, 'TSTOP',  stopTime)

    if (numLiveSources > 0) then
      srcNumCol    = addColumn(tab, 'SRC_NUM',     INTEGER32)
      raCol        = addColumn(tab, 'RA',          REAL64,    'deg')
      decCol       = addColumn(tab, 'DEC',         REAL64,    'deg')
      raCorrCol    = addColumn(tab, 'RA_CORR',     REAL64,    'deg')
      decCorrCol   = addColumn(tab, 'DEC_CORR',    REAL64,    'deg')
      binWidthCol  = addColumn(tab, 'TBIN_WIDTH',  REAL64,    's')
      srcNumHexCol = addColumn(tab, 'SRC_NUM_HEX', STRING&
        , dimensions=(/maxNumHexSrcChars/))
      bitFlagsCol  = addColumn(tab, 'PROCESS',     INTEGER8&
        , comment='If the relevant bit is set, make products for this src.')
      epicSctsCol  = addColumn(tab, 'EP_CTS',      REAL32,    'counts'&
        , comment='Equivalent epic source counts')
      pnSctsCol    = addColumn(tab, 'PN_CTS',      REAL32,    'counts')
      m1SctsCol    = addColumn(tab, 'M1_CTS',      REAL32,    'counts')
      m2SctsCol    = addColumn(tab, 'M2_CTS',      REAL32,    'counts')
      mlIdSrcCol   = addColumn(tab, 'ML_ID_SRC',   INTEGER32)
      epFlagCol    = addColumn(tab, 'EP_FLAG',     STRING, dimensions=(/LENGTHOFINSTFLAG/))
      pnFlagCol    = addColumn(tab, 'PN_FLAG',     STRING, dimensions=(/LENGTHOFINSTFLAG/))
      m1FlagCol    = addColumn(tab, 'M1_FLAG',     STRING, dimensions=(/LENGTHOFINSTFLAG/))
      m2FlagCol    = addColumn(tab, 'M2_FLAG',     STRING, dimensions=(/LENGTHOFINSTFLAG/))

      row = 0
      do i = 1, numSources
      if (srcData(i)%bitFlags == 0) cycle

        if (srcData(i)%sctsIsNull)&
          call setToNull(epicSctsCol, row)
        if (srcData(i)%PN%sctsIsNull)&
          call setToNull(pnSctsCol, row)
        if (srcData(i)%M1%sctsIsNull)&
          call setToNull(m1SctsCol, row)
        if (srcData(i)%M2%sctsIsNull)&
          call setToNull(m2SctsCol, row)
        if (srcData(i)%mlIdSrc <= 0)&
          call setToNull(mlIdSrcCol, row)

        row = row + 1
      end do

      srcNumData    => int32Data(   srcNumCol)
      raData        => real64Data(      raCol)
      decData       => real64Data(     decCol)
      raCorrData    => real64Data(  raCorrCol)
      decCorrData   => real64Data( decCorrCol)
      binWidthData  => real64Data(binWidthCol)
      bitFlagsData  => int8Data(  bitFlagsCol)
      epicSctsData  => real32Data(epicSctsCol)
      pnSctsData    => real32Data(  pnSctsCol)
      m1SctsData    => real32Data(  m1SctsCol)
      m2SctsData    => real32Data(  m2SctsCol)
      mlIdSrcData   => int32Data(  mlIdSrcCol)

      row = 0
      do i = 1, numSources
      if (srcData(i)%bitFlags == 0) cycle

        srcNumData(  row) = srcData(i)%srcNum
        raData(      row) = srcData(i)%ra
        decData(     row) = srcData(i)%dec
        raCorrData(  row) = srcData(i)%raCorr
        decCorrData( row) = srcData(i)%decCorr
        binWidthData(row) = srcData(i)%binWidth
        bitFlagsData(row) = srcData(i)%bitFlags

        ! Note: Do not trim() the input string, as it will give you a junk at the end of each cell in the FITS.
        call setStringCell(epFlagCol, row, srcData(i)%epFlag)	! in DAL
        call setStringCell(pnFlagCol, row, srcData(i)%pnFlag)
        call setStringCell(m1FlagCol, row, srcData(i)%m1Flag)
        call setStringCell(m2FlagCol, row, srcData(i)%m2Flag)

!        call setStringCell(srcNumHexCol, row, trim(srcData(i)%srcNumHex))
        call setStringCell(srcNumHexCol, row, srcData(i)%srcNumHex)
!*** check that this truncates ok

        if (.not. srcData(i)%sctsIsNull)&
          epicSctsData(row) = srcData(i)%scts
        if (.not. srcData(i)%PN%sctsIsNull)&
          pnSctsData(row) = srcData(i)%PN%scts
        if (.not. srcData(i)%M1%sctsIsNull)&
          m1SctsData(row) = srcData(i)%M1%scts
        if (.not. srcData(i)%M2%sctsIsNull)&
          m2SctsData(row) = srcData(i)%M2%scts
        if (srcData(i)%mlIdSrc > 0)&
          mlIdSrcData(row) = srcData(i)%mlIdSrc

        row = row + 1
      end do

      call release(m2FlagCol)
      call release(m1FlagCol)
      call release(pnFlagCol)
      call release(epFlagCol)
      call release(mlIdSrcCol)
      call release(m2SctsCol)
      call release(m1SctsCol)
      call release(pnSctsCol)
      call release(epicSctsCol)
      call release(bitFlagsCol)
      call release(srcNumHexCol)
      call release(binWidthCol)
      call release(decCol)
      call release(raCol)
      call release(srcNumCol)
    end if

    call release(tab)
    call release(set)

  end subroutine writeSourceListData

  !---------------------------------------------------------------------
  subroutine setFlags(srcIsSelected)
  !---------------------------------------------------------------------

    implicit none

    logical, intent(in) :: srcIsSelected(:)

    type(DataSetT) :: set
    type(TableT) :: tab
    type(ColumnT) :: col0, col1
    logical(bool), pointer :: spectraFlag(:), tSeriesFlag(:)

    !-------------------------------------------------------------------
    ! Start

    set = dataSet(srcmatchSourceSetName, MODIFY)
    tab = table(set, srcmatchSourceTabName)
    col0 = column(tab, 'SPECTRA')
    col1 = column(tab, 'TSERIES')
    spectraFlag => boolData(col0)
    tSeriesFlag => boolData(col1)

    spectraFlag = .false.
    tSeriesFlag = .false.
    where(srcIsSelected)
      spectraFlag = .true.
      tSeriesFlag = .true.
    end where

    call release(col1)
    call release(col0)
    call release(tab)
    call release(set)

  end subroutine setFlags

  !---------------------------------------------------------------------
  !function intToHex(intArg) result(hexStr)
  !---------------------------------------------------------------------
  ! Converts the integer into a left-justified hexadecimal string.
  !
  !end function intToHex


end module esources_mod
